@@ -18,6 +18,7 @@ func (s *aztestsSuite) TestGetAccountInfo(c *chk.C) {
 
 	// Test on a container
 	cURL := sa.NewContainerURL(generateContainerName())
+	defer deleteContainer(c, cURL, false)
 	_, err = cURL.Create(ctx, Metadata{}, PublicAccessNone)
 	c.Assert(err, chk.IsNil)
 	cAccInfo, err := cURL.GetAccountInfo(ctx)
@@ -26,7 +27,7 @@ func (s *aztestsSuite) TestGetAccountInfo(c *chk.C) {
 
 	// test on a block blob URL. They all call the same thing on the base URL, so only one test is needed for that.
 	bbURL := cURL.NewBlockBlobURL(generateBlobName())
-	_, err = bbURL.Upload(ctx, strings.NewReader("blah"), BlobHTTPHeaders{}, Metadata{}, BlobAccessConditions{})
+	_, err = bbURL.Upload(ctx, strings.NewReader("blah"), BlobHTTPHeaders{}, Metadata{}, BlobAccessConditions{}, DefaultAccessTier, nil, ClientProvidedKeyOptions{}, ImmutabilityPolicyOptions{})
 	c.Assert(err, chk.IsNil)
 	bAccInfo, err := bbURL.GetAccountInfo(ctx)
 	c.Assert(err, chk.IsNil)
@@ -44,7 +45,7 @@ func (s *aztestsSuite) TestListContainers(c *chk.C) {
 	c.Assert(resp.ServiceEndpoint, chk.NotNil)
 
 	container, name := createNewContainer(c, sa)
-	defer delContainer(c, container)
+	defer deleteContainer(c, container, false)
 
 	md := Metadata{
 		"foo": "foovalue",
@@ -81,7 +82,7 @@ func (s *aztestsSuite) TestListContainersPaged(c *chk.C) {
 
 	defer func() {
 		for i := range containers {
-			delContainer(c, containers[i])
+			deleteContainer(c, containers[i], false)
 		}
 	}()
 
@@ -102,9 +103,9 @@ func (s *aztestsSuite) TestListContainersPaged(c *chk.C) {
 func (s *aztestsSuite) TestAccountListContainersEmptyPrefix(c *chk.C) {
 	bsu := getBSU()
 	containerURL1, _ := createNewContainer(c, bsu)
-	defer deleteContainer(c, containerURL1)
+	defer deleteContainer(c, containerURL1, false)
 	containerURL2, _ := createNewContainer(c, bsu)
-	defer deleteContainer(c, containerURL2)
+	defer deleteContainer(c, containerURL2, false)
 
 	response, err := bsu.ListContainersSegment(ctx, Marker{}, ListContainersSegmentOptions{})
 
@@ -115,9 +116,9 @@ func (s *aztestsSuite) TestAccountListContainersEmptyPrefix(c *chk.C) {
 func (s *aztestsSuite) TestAccountListContainersIncludeTypeMetadata(c *chk.C) {
 	bsu := getBSU()
 	containerURLNoMetadata, nameNoMetadata := createNewContainerWithSuffix(c, bsu, "a")
-	defer deleteContainer(c, containerURLNoMetadata)
+	defer deleteContainer(c, containerURLNoMetadata, false)
 	containerURLMetadata, nameMetadata := createNewContainerWithSuffix(c, bsu, "b")
-	defer deleteContainer(c, containerURLMetadata)
+	defer deleteContainer(c, containerURLMetadata, false)
 
 	// Test on containers with and without metadata
 	_, err := containerURLMetadata.SetMetadata(ctx, basicMetadata, ContainerAccessConditions{})
@@ -137,7 +138,7 @@ func (s *aztestsSuite) TestAccountListContainersMaxResultsNegative(c *chk.C) {
 	bsu := getBSU()
 	containerURL, _ := createNewContainer(c, bsu)
 
-	defer deleteContainer(c, containerURL)
+	defer deleteContainer(c, containerURL, false)
 	_, err := bsu.ListContainersSegment(ctx,
 		Marker{}, *(&ListContainersSegmentOptions{Prefix: containerPrefix, MaxResults: -2}))
 	c.Assert(err, chk.Not(chk.IsNil))
@@ -147,7 +148,7 @@ func (s *aztestsSuite) TestAccountListContainersMaxResultsZero(c *chk.C) {
 	bsu := getBSU()
 	containerURL, _ := createNewContainer(c, bsu)
 
-	defer deleteContainer(c, containerURL)
+	defer deleteContainer(c, containerURL, false)
 
 	// Max Results = 0 means the value will be ignored, the header not set, and the server default used
 	resp, err := bsu.ListContainersSegment(ctx,
@@ -162,9 +163,9 @@ func (s *aztestsSuite) TestAccountListContainersMaxResultsExact(c *chk.C) {
 	// If this test fails, ensure there are no extra containers prefixed with go in the account. These may be left over if a test is interrupted.
 	bsu := getBSU()
 	containerURL1, containerName1 := createNewContainerWithSuffix(c, bsu, "a")
-	defer deleteContainer(c, containerURL1)
+	defer deleteContainer(c, containerURL1, false)
 	containerURL2, containerName2 := createNewContainerWithSuffix(c, bsu, "b")
-	defer deleteContainer(c, containerURL2)
+	defer deleteContainer(c, containerURL2, false)
 
 	response, err := bsu.ListContainersSegment(ctx,
 		Marker{}, *(&ListContainersSegmentOptions{Prefix: containerPrefix, MaxResults: 2}))
@@ -178,9 +179,9 @@ func (s *aztestsSuite) TestAccountListContainersMaxResultsExact(c *chk.C) {
 func (s *aztestsSuite) TestAccountListContainersMaxResultsInsufficient(c *chk.C) {
 	bsu := getBSU()
 	containerURL1, _ := createNewContainer(c, bsu)
-	defer deleteContainer(c, containerURL1)
+	defer deleteContainer(c, containerURL1, false)
 	containerURL2, _ := createNewContainer(c, bsu)
-	defer deleteContainer(c, containerURL2)
+	defer deleteContainer(c, containerURL2, false)
 
 	response, err := bsu.ListContainersSegment(ctx, Marker{},
 		*(&ListContainersSegmentOptions{Prefix: containerPrefix, MaxResults: 1}))
@@ -192,9 +193,9 @@ func (s *aztestsSuite) TestAccountListContainersMaxResultsInsufficient(c *chk.C)
 func (s *aztestsSuite) TestAccountListContainersMaxResultsSufficient(c *chk.C) {
 	bsu := getBSU()
 	containerURL1, _ := createNewContainer(c, bsu)
-	defer deleteContainer(c, containerURL1)
+	defer deleteContainer(c, containerURL1, false)
 	containerURL2, _ := createNewContainer(c, bsu)
-	defer deleteContainer(c, containerURL2)
+	defer deleteContainer(c, containerURL2, false)
 
 	response, err := bsu.ListContainersSegment(ctx, Marker{},
 		*(&ListContainersSegmentOptions{Prefix: containerPrefix, MaxResults: 3}))
@@ -205,6 +206,177 @@ func (s *aztestsSuite) TestAccountListContainersMaxResultsSufficient(c *chk.C) {
 	c.Assert(len(response.ContainerItems) >= 2, chk.Equals, true)
 }
 
+func CreateBlobWithRetentionPolicy(c *chk.C) (BlockBlobURL, ContainerURL) {
+	bsu, err := getAlternateBSU()
+	c.Assert(err, chk.IsNil)
+
+	days := int32(5)
+	allowDelete := true
+	_, err = bsu.SetProperties(ctx, StorageServiceProperties{DeleteRetentionPolicy: &RetentionPolicy{Enabled: true, Days: &days, AllowPermanentDelete: &allowDelete}})
+	c.Assert(err, chk.IsNil)
+
+	// From FE, 30 seconds is guaranteed to be enough.
+	time.Sleep(time.Second * 30)
+
+	// create container and blobs
+	containerURL, _ := createNewContainer(c, bsu)
+	testSize := 8 * 1024
+	r, _ := getRandomDataAndReader(testSize)
+	blobURL, _ := getBlockBlobURL(c, containerURL)
+
+	cResp, err := blobURL.Upload(ctx, r, BlobHTTPHeaders{}, nil, BlobAccessConditions{}, DefaultAccessTier, nil, ClientProvidedKeyOptions{}, ImmutabilityPolicyOptions{})
+	c.Assert(err, chk.IsNil)
+	c.Assert(cResp.StatusCode(), chk.Equals, 201)
+
+	return blobURL, containerURL
+}
+
+func (s *aztestsSuite) TestUndelete(c *chk.C) {
+	blobURL, containerURL := CreateBlobWithRetentionPolicy(c)
+	defer deleteContainer(c, containerURL, false)
+
+	// Soft delete blob
+	_, err := blobURL.Delete(ctx, DeleteSnapshotsOptionNone, BlobAccessConditions{}) //soft delete
+	c.Assert(err, chk.IsNil)
+
+	// Check that blob has been soft deleted
+	_, err = blobURL.GetProperties(ctx, BlobAccessConditions{}, ClientProvidedKeyOptions{})
+	c.Assert(err, chk.NotNil)
+
+	// Undelete soft deleted blob
+	undelResp, err := blobURL.Undelete(ctx)
+	c.Assert(err, chk.IsNil)
+	c.Assert(undelResp, chk.NotNil)
+	c.Assert(undelResp.StatusCode(), chk.Equals, 200)
+
+	blobProp, err := blobURL.GetProperties(ctx, BlobAccessConditions{}, ClientProvidedKeyOptions{})
+	c.Assert(blobProp.StatusCode(), chk.Equals, 200)
+}
+
+/*func (s *aztestsSuite) TestPermanentDelete(c *chk.C) {
+	blobURL, containerURL := CreateBlobWithRetentionPolicy(c)
+	defer deleteContainer(c, containerURL, false)
+
+	// Create snapshot for blob
+	snapResp, err := blobURL.CreateSnapshot(ctx, Metadata{}, BlobAccessConditions{}, ClientProvidedKeyOptions{})
+	c.Assert(snapResp, chk.NotNil)
+	c.Assert(err, chk.IsNil)
+
+	// Check snapshot and blob exist
+	listBlobResp1, err := containerURL.ListBlobsFlatSegment(ctx, Marker{}, ListBlobsSegmentOptions{Details: BlobListingDetails{Snapshots: true}})
+	c.Assert(err, chk.IsNil)
+	c.Assert(listBlobResp1.Segment.BlobItems, chk.HasLen, 2)
+
+	// Soft delete snapshot
+	snapshotBlob := blobURL.WithSnapshot(snapResp.Snapshot())
+	_, err = snapshotBlob.Delete(ctx, DeleteSnapshotsOptionNone, BlobAccessConditions{})
+	c.Assert(err, chk.IsNil)
+
+	// Check that both blobs and snapshot exist
+	listBlobResp2, err := containerURL.ListBlobsFlatSegment(ctx, Marker{}, ListBlobsSegmentOptions{Details: BlobListingDetails{Deleted: true, Snapshots: true}})
+	c.Assert(err, chk.IsNil)
+	c.Assert(listBlobResp2.Segment.BlobItems, chk.HasLen, 2)
+
+	// Permanent delete snapshot
+	delResp, err := snapshotBlob.PermanentDelete(ctx, DeleteSnapshotsOptionNone, BlobAccessConditions{})
+	c.Assert(err, chk.IsNil)
+	c.Assert(delResp, chk.NotNil)
+	c.Assert(delResp.StatusCode(), chk.Equals, 202)
+
+	// Check that snapshot has been deleted
+	spBlobResp, err := snapshotBlob.GetProperties(ctx, BlobAccessConditions{}, ClientProvidedKeyOptions{})
+	c.Assert(err, chk.NotNil)
+	c.Assert(spBlobResp, chk.IsNil)
+
+	// Check that only blob exists
+	listBlobResp3, err := containerURL.ListBlobsFlatSegment(ctx, Marker{}, ListBlobsSegmentOptions{Details: BlobListingDetails{Deleted: true, Snapshots: true}})
+	c.Assert(err, chk.IsNil)
+	c.Assert(listBlobResp3.Segment.BlobItems, chk.HasLen, 1)
+	c.Assert(listBlobResp3.Segment.BlobItems[0].Deleted, chk.Equals, false)
+}*/
+
+/*func (s *aztestsSuite) TestPermanentDeleteAccountSAS(c *chk.C) {
+	accountName, accountKey := accountInfo()
+	credential, err := NewSharedKeyCredential(accountName, accountKey)
+	if err != nil {
+		c.Fail()
+	}
+
+	sasQueryParams, err := AccountSASSignatureValues{
+		Protocol:      SASProtocolHTTPS,
+		ExpiryTime:    time.Now().UTC().Add(48 * time.Hour),
+		Permissions:   AccountSASPermissions{Read: true, List: true, Write: true, Create: true, PermanentDelete: true, Delete: true}.String(),
+		Services:      AccountSASServices{Blob: true}.String(),
+		ResourceTypes: AccountSASResourceTypes{Service: true, Container: true, Object: true}.String(),
+	}.NewSASQueryParameters(credential)
+	if err != nil {
+		log.Fatal(err)
+	}
+
+	qp := sasQueryParams.Encode()
+	urlToSendToSomeone := fmt.Sprintf("https://%s.blob.core.windows.net?%s", accountName, qp)
+	u, _ := url.Parse(urlToSendToSomeone)
+	serviceURL := NewServiceURL(*u, NewPipeline(NewAnonymousCredential(), PipelineOptions{}))
+	days := int32(5)
+	allowDelete := true
+	_, err = serviceURL.SetProperties(ctx, StorageServiceProperties{DeleteRetentionPolicy: &RetentionPolicy{Enabled: true, Days: &days, AllowPermanentDelete: &allowDelete}})
+	c.Assert(err, chk.IsNil)
+
+	// From FE, 30 seconds is guaranteed to be enough.
+	time.Sleep(time.Second * 30)
+
+	containerName := generateContainerName()
+	containerURL := serviceURL.NewContainerURL(containerName)
+	_, err = containerURL.Create(ctx, Metadata{}, PublicAccessNone)
+	defer containerURL.Delete(ctx, ContainerAccessConditions{})
+	if err != nil {
+		c.Fatal(err)
+	}
+
+	blobURL := containerURL.NewBlockBlobURL("temp")
+	_, err = blobURL.Upload(ctx, bytes.NewReader([]byte("random data")), BlobHTTPHeaders{}, basicMetadata, BlobAccessConditions{}, DefaultAccessTier, nil, ClientProvidedKeyOptions{})
+	if err != nil {
+		c.Fail()
+	}
+
+	// Create snapshot for blob
+	snapResp, err := blobURL.CreateSnapshot(ctx, Metadata{}, BlobAccessConditions{}, ClientProvidedKeyOptions{})
+	c.Assert(snapResp, chk.NotNil)
+	c.Assert(err, chk.IsNil)
+
+	// Check snapshot and blob exist
+	listBlobResp1, err := containerURL.ListBlobsFlatSegment(ctx, Marker{}, ListBlobsSegmentOptions{Details: BlobListingDetails{Snapshots: true}})
+	c.Assert(err, chk.IsNil)
+	c.Assert(listBlobResp1.Segment.BlobItems, chk.HasLen, 2)
+
+	// Soft delete snapshot
+	snapshotBlob := blobURL.WithSnapshot(snapResp.Snapshot())
+	_, err = snapshotBlob.Delete(ctx, DeleteSnapshotsOptionNone, BlobAccessConditions{})
+	c.Assert(err, chk.IsNil)
+
+	// Check that both blobs and snapshot exist
+	listBlobResp2, err := containerURL.ListBlobsFlatSegment(ctx, Marker{}, ListBlobsSegmentOptions{Details: BlobListingDetails{Deleted: true, Snapshots: true}})
+	c.Assert(err, chk.IsNil)
+	c.Assert(listBlobResp2.Segment.BlobItems, chk.HasLen, 2)
+
+	// Permanent delete snapshot
+	delResp, err := snapshotBlob.PermanentDelete(ctx, DeleteSnapshotsOptionNone, BlobAccessConditions{})
+	c.Assert(err, chk.IsNil)
+	c.Assert(delResp, chk.NotNil)
+	c.Assert(delResp.StatusCode(), chk.Equals, 202)
+
+	// Check that snapshot has been deleted
+	spBlobResp, err := snapshotBlob.GetProperties(ctx, BlobAccessConditions{}, ClientProvidedKeyOptions{})
+	c.Assert(err, chk.NotNil)
+	c.Assert(spBlobResp, chk.IsNil)
+
+	// Check that only blob exists
+	listBlobResp3, err := containerURL.ListBlobsFlatSegment(ctx, Marker{}, ListBlobsSegmentOptions{Details: BlobListingDetails{Deleted: true, Snapshots: true}})
+	c.Assert(err, chk.IsNil)
+	c.Assert(listBlobResp3.Segment.BlobItems, chk.HasLen, 1)
+	c.Assert(listBlobResp3.Segment.BlobItems[0].Deleted, chk.Equals, false)
+}*/
+
 func (s *aztestsSuite) TestAccountDeleteRetentionPolicy(c *chk.C) {
 	bsu := getBSU()
 
